=== modified file 'TODO'
--- TODO	2010-06-19 00:37:04 +0000
+++ TODO	2010-09-02 18:53:38 +0000
@@ -83,6 +83,9 @@
 ** TODO [#C] python-parsedatetime
 ** TODO [#C] systemd/launchd
    http://0pointer.de/blog/projects/systemd.html
+** TODO Separate logging logic to own object
+** TODO make clients to a dict!
+** TODO PropertyChanged should be a python generic property for client variables
 
 * mandos.xml
 ** [[file:mandos.xml::XXX][Document D-Bus interface]]
@@ -126,5 +129,9 @@
 ** TODO [#C] /etc/bash_completion.d/mandos
    From XML sources directly?
 
+* Side Stuff
+** TODO Locate which packet move the other bin/sh when busy box is deactivated
+** TODO contact owner of packet, and ask them to have that shell static in position regardless of busybox
+
 
 #+STARTUP: showall

=== modified file 'mandos'
--- mandos	2010-06-19 00:37:04 +0000
+++ mandos	2010-09-02 18:53:38 +0000
@@ -60,7 +60,7 @@
 import fcntl
 import functools
 import cPickle as pickle
-import select
+import multiprocessing
 
 import dbus
 import dbus.service
@@ -258,6 +258,9 @@
                      runtime with vars(self) as dict, so that for
                      instance %(name)s can be used in the command.
     current_checker_command: string; current running checker_command
+    approved_delay: datetime.timedelta(); Time to wait for approval
+    _approved:   bool(); 'None' if not yet approved/disapproved
+    approved_duration: datetime.timedelta(); Duration of one approval
     """
     
     @staticmethod
@@ -274,6 +277,9 @@
     def interval_milliseconds(self):
         "Return the 'interval' attribute in milliseconds"
         return self._timedelta_to_milliseconds(self.interval)
+
+    def approved_delay_milliseconds(self):
+        return self._timedelta_to_milliseconds(self.approved_delay)
     
     def __init__(self, name = None, disable_hook=None, config=None):
         """Note: the 'checker' key in 'config' sets the
@@ -297,6 +303,7 @@
                       "rb") as secfile:
                 self.secret = secfile.read()
         else:
+            #XXX Need to allow secret on demand!
             raise TypeError(u"No secret or secfile for client %s"
                             % self.name)
         self.host = config.get(u"host", u"")
@@ -314,12 +321,27 @@
         self.checker_command = config[u"checker"]
         self.current_checker_command = None
         self.last_connect = None
-    
+        self.approvals_pending = 0
+        self._approved = None
+        self.approved_by_default = config.get(u"approved_by_default",
+                                              False)
+        self.approved_delay = string_to_delta(
+            config[u"approved_delay"])
+        self.approved_duration = string_to_delta(
+            config[u"approved_duration"])
+        self.changedstate = multiprocessing_manager.Condition(multiprocessing_manager.Lock())
+
+    def send_changedstate(self):
+        self.changedstate.acquire()
+        self.changedstate.notify_all()
+        self.changedstate.release()
+        
     def enable(self):
         """Start this client's checker and timeout hooks"""
         if getattr(self, u"enabled", False):
             # Already enabled
             return
+        self.send_changedstate()
         self.last_enabled = datetime.datetime.utcnow()
         # Schedule a new checker to be started an 'interval' from now,
         # and every interval from then on.
@@ -339,6 +361,8 @@
         if not getattr(self, "enabled", False):
             return False
         if not quiet:
+            self.send_changedstate()
+        if not quiet:
             logger.info(u"Disabling client %s", self.name)
         if getattr(self, u"disable_initiator_tag", False):
             gobject.source_remove(self.disable_initiator_tag)
@@ -477,7 +501,6 @@
                 raise
         self.checker = None
 
-
 def dbus_service_property(dbus_interface, signature=u"v",
                           access=u"readwrite", byte_arrays=False):
     """Decorators for marking methods of a DBusObjectWithProperties to
@@ -780,6 +803,18 @@
             self.PropertyChanged(dbus.String(u"checker_running"),
                                  dbus.Boolean(False, variant_level=1))
         return r
+
+    def _reset_approved(self):
+        self._approved = None
+        return False
+    
+    def approve(self, value=True):
+        self._approved = value
+        gobject.timeout_add(self._timedelta_to_milliseconds(self.approved_duration, self._reset_approved))
+
+    def approved_pending(self):
+        return self.approvals_pending > 0
+
     
     ## D-Bus methods, signals & properties
     _interface = u"se.bsnet.fukt.Mandos.Client"
@@ -805,19 +840,31 @@
         pass
     
     # GotSecret - signal
+    # Is sent after succesfull transfer of secret from mandos-server to mandos-client
     @dbus.service.signal(_interface)
     def GotSecret(self):
         "D-Bus signal"
         pass
     
     # Rejected - signal
-    @dbus.service.signal(_interface)
-    def Rejected(self):
+    @dbus.service.signal(_interface, signature=u"s")
+    def Rejected(self, reason):
+        "D-Bus signal"
+        pass
+    
+    # NeedApproval - signal
+    @dbus.service.signal(_interface, signature=u"db")
+    def NeedApproval(self, timeout, default):
         "D-Bus signal"
         pass
     
     ## Methods
-    
+
+    # Approve - method
+    @dbus.service.method(_interface, in_signature=u"b")
+    def Approve(self, value):
+        self.approve(value)
+
     # CheckedOK - method
     @dbus.service.method(_interface)
     def CheckedOK(self):
@@ -848,6 +895,30 @@
     
     ## Properties
     
+    # approved_pending - property
+    @dbus_service_property(_interface, signature=u"b", access=u"read")
+    def approved_pending_dbus_property(self):
+        return dbus.Boolean(self.approved_pending())
+    
+    # approved_by_default - property
+    @dbus_service_property(_interface, signature=u"b",
+                           access=u"readwrite")
+    def approved_by_default_dbus_property(self):
+        return dbus.Boolean(self.approved_by_default)
+    
+    # approved_delay - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def approved_delay_dbus_property(self):
+        return dbus.UInt64(self.approved_delay_milliseconds())
+    
+    # approved_duration - property
+    @dbus_service_property(_interface, signature=u"t",
+                           access=u"readwrite")
+    def approved_duration_dbus_property(self):
+        return dbus.UInt64(self._timedelta_to_milliseconds(
+                self.approved_duration))
+    
     # name - property
     @dbus_service_property(_interface, signature=u"s", access=u"read")
     def name_dbus_property(self):
@@ -987,6 +1058,32 @@
     del _interface
 
 
+class ProxyClient(object):
+    def __init__(self, child_pipe, fpr, address):
+        self._pipe = child_pipe
+        self._pipe.send(('init', fpr, address))
+        if not self._pipe.recv():
+            raise KeyError()
+
+    def __getattribute__(self, name):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__getattribute__(name)
+        self._pipe.send(('getattr', name))
+        data = self._pipe.recv()
+        if data[0] == 'data':
+            return data[1]
+        if data[0] == 'function':
+            def func(*args, **kwargs):
+                self._pipe.send(('funcall', name, args, kwargs))
+                return self._pipe.recv()[1]
+            return func
+
+    def __setattr__(self, name, value):
+        if(name == '_pipe'):
+            return super(ProxyClient, self).__setattr__(name, value)
+        self._pipe.send(('setattr', name, value))
+
+
 class ClientHandler(socketserver.BaseRequestHandler, object):
     """A class to handle client connections.
     
@@ -994,24 +1091,22 @@
     Note: This will run in its own forked process."""
     
     def handle(self):
-        logger.info(u"TCP connection from: %s",
-                    unicode(self.client_address))
-        logger.debug(u"IPC Pipe FD: %d",
-                     self.server.child_pipe[1].fileno())
-        # Open IPC pipe to parent process
-        with contextlib.nested(self.server.child_pipe[1],
-                               self.server.parent_pipe[0]
-                               ) as (ipc, ipc_return):
+        with contextlib.closing(self.server.child_pipe) as child_pipe:
+            logger.info(u"TCP connection from: %s",
+                        unicode(self.client_address))
+            logger.debug(u"Pipe FD: %d",
+                         self.server.child_pipe.fileno())
+
             session = (gnutls.connection
                        .ClientSession(self.request,
                                       gnutls.connection
                                       .X509Credentials()))
-            
+
             # Note: gnutls.connection.X509Credentials is really a
             # generic GnuTLS certificate credentials object so long as
             # no X.509 keys are added to it.  Therefore, we can use it
             # here despite using OpenPGP certificates.
-            
+
             #priority = u':'.join((u"NONE", u"+VERS-TLS1.1",
             #                      u"+AES-256-CBC", u"+SHA1",
             #                      u"+COMP-NULL", u"+CTYPE-OPENPGP",
@@ -1023,7 +1118,7 @@
             (gnutls.library.functions
              .gnutls_priority_set_direct(session._c_object,
                                          priority, None))
-            
+
             # Start communication using the Mandos protocol
             # Get protocol number
             line = self.request.makefile().readline()
@@ -1034,7 +1129,7 @@
             except (ValueError, IndexError, RuntimeError), error:
                 logger.error(u"Unknown protocol version: %s", error)
                 return
-            
+
             # Start GnuTLS connection
             try:
                 session.handshake()
@@ -1044,6 +1139,8 @@
                 # established.  Just abandon the request.
                 return
             logger.debug(u"Handshake succeeded")
+
+            approval_required = False
             try:
                 try:
                     fpr = self.fingerprint(self.peer_certificate
@@ -1053,47 +1150,85 @@
                     return
                 logger.debug(u"Fingerprint: %s", fpr)
 
-                for c in self.server.clients:
-                    if c.fingerprint == fpr:
-                        client = c
+                try:
+                    client = ProxyClient(child_pipe, fpr,
+                                         self.client_address)
+                except KeyError:
+                    return
+                
+                if client.approved_delay:
+                    delay = client.approved_delay
+                    client.approvals_pending += 1
+                    approval_required = True
+                
+                while True:
+                    if not client.enabled:
+                        logger.warning(u"Client %s is disabled",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Disabled")                    
+                        return
+                    
+                    if client._approved or not client.approved_delay:
+                        #We are approved or approval is disabled
                         break
-                else:
-                    ipc.write(u"NOTFOUND %s %s\n"
-                              % (fpr, unicode(self.client_address)))
-                    return
-                
-                class ClientProxy(object):
-                    """Client proxy object.  Not for calling methods."""
-                    def __init__(self, client):
-                        self.client = client
-                    def __getattr__(self, name):
-                        if name.startswith("ipc_"):
-                            def tempfunc():
-                                ipc.write("%s %s\n" % (name[4:].upper(),
-                                                       self.client.name))
-                            return tempfunc
-                        if not hasattr(self.client, name):
-                            raise AttributeError
-                        ipc.write(u"GETATTR %s %s\n"
-                                  % (name, self.client.fingerprint))
-                        return pickle.load(ipc_return)
-                clientproxy = ClientProxy(client)
-                # Have to check if client.enabled, since it is
-                # possible that the client was disabled since the
-                # GnuTLS session was established.
-                if not clientproxy.enabled:
-                    clientproxy.ipc_disabled()
-                    return
-                
-                clientproxy.ipc_sending()
+                    elif client._approved is None:
+                        logger.info(u"Client %s need approval",
+                                    client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.NeedApproval(
+                                client.approved_delay_milliseconds(),
+                                client.approved_by_default)
+                    else:
+                        logger.warning(u"Client %s was not approved",
+                                       client.name)
+                        if self.server.use_dbus:
+                            # Emit D-Bus signal
+                            client.Rejected("Disapproved")
+                        return
+                    
+                    #wait until timeout or approved
+                    #x = float(client._timedelta_to_milliseconds(delay))
+                    time = datetime.datetime.now()
+                    client.changedstate.acquire()
+                    client.changedstate.wait(float(client._timedelta_to_milliseconds(delay) / 1000))
+                    client.changedstate.release()
+                    time2 = datetime.datetime.now()
+                    if (time2 - time) >= delay:
+                        if not client.approved_by_default:
+                            logger.warning("Client %s timed out while"
+                                           " waiting for approval",
+                                           client.name)
+                            if self.server.use_dbus:
+                                # Emit D-Bus signal
+                                client.Rejected("Time out")
+                            return
+                        else:
+                            break
+                    else:
+                        delay -= time2 - time
+                
                 sent_size = 0
                 while sent_size < len(client.secret):
+                    # XXX handle session exception
                     sent = session.send(client.secret[sent_size:])
                     logger.debug(u"Sent: %d, remaining: %d",
                                  sent, len(client.secret)
                                  - (sent_size + sent))
                     sent_size += sent
+
+                logger.info(u"Sending secret to %s", client.name)
+                # bump the timeout as if seen
+                client.checked_ok()
+                if self.server.use_dbus:
+                    # Emit D-Bus signal
+                    client.GotSecret()
+            
             finally:
+                if approval_required:
+                    client.approvals_pending -= 1
                 session.bye()
     
     @staticmethod
@@ -1160,30 +1295,39 @@
         return hex_fpr
 
 
-class ForkingMixInWithPipes(socketserver.ForkingMixIn, object):
-    """Like socketserver.ForkingMixIn, but also pass a pipe pair."""
+class MultiprocessingMixIn(object):
+    """Like socketserver.ThreadingMixIn, but with multiprocessing"""
+    def sub_process_main(self, request, address):
+        try:
+            self.finish_request(request, address)
+        except:
+            self.handle_error(request, address)
+        self.close_request(request)
+            
+    def process_request(self, request, address):
+        """Start a new process to process the request."""
+        multiprocessing.Process(target = self.sub_process_main,
+                                args = (request, address)).start()
+
+class MultiprocessingMixInWithPipe(MultiprocessingMixIn, object):
+    """ adds a pipe to the MixIn """
     def process_request(self, request, client_address):
         """Overrides and wraps the original process_request().
         
         This function creates a new pipe in self.pipe
         """
-        # Child writes to child_pipe
-        self.child_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        # Parent writes to parent_pipe
-        self.parent_pipe = map(os.fdopen, os.pipe(), u"rw", (1, 0))
-        super(ForkingMixInWithPipes,
+        parent_pipe, self.child_pipe = multiprocessing.Pipe()
+
+        super(MultiprocessingMixInWithPipe,
               self).process_request(request, client_address)
-        # Close unused ends for parent
-        self.parent_pipe[0].close() # close read end
-        self.child_pipe[1].close()  # close write end
-        self.add_pipe_fds(self.child_pipe[0], self.parent_pipe[1])
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+        self.child_pipe.close()
+        self.add_pipe(parent_pipe)
+        
+    def add_pipe(self, parent_pipe):
         """Dummy function; override as necessary"""
-        child_pipe_fd.close()
-        parent_pipe_fd.close()
-
-
-class IPv6_TCPServer(ForkingMixInWithPipes,
+        pass
+
+class IPv6_TCPServer(MultiprocessingMixInWithPipe,
                      socketserver.TCPServer, object):
     """IPv6-capable TCP server.  Accepts 'None' as address and/or port
     
@@ -1274,14 +1418,15 @@
             return socketserver.TCPServer.server_activate(self)
     def enable(self):
         self.enabled = True
-    def add_pipe_fds(self, child_pipe_fd, parent_pipe_fd):
+    def add_pipe(self, parent_pipe):
         # Call "handle_ipc" for both data and EOF events
-        gobject.io_add_watch(child_pipe_fd.fileno(),
+        gobject.io_add_watch(parent_pipe.fileno(),
                              gobject.IO_IN | gobject.IO_HUP,
                              functools.partial(self.handle_ipc,
-                                               reply = parent_pipe_fd,
-                                               sender= child_pipe_fd))
-    def handle_ipc(self, source, condition, reply=None, sender=None):
+                                               parent_pipe = parent_pipe))
+        
+    def handle_ipc(self, source, condition, parent_pipe=None,
+                   client_object=None):
         condition_names = {
             gobject.IO_IN: u"IN",   # There is data to read.
             gobject.IO_OUT: u"OUT", # Data can be written (without
@@ -1298,68 +1443,60 @@
                                        if cond & condition)
         logger.debug(u"Handling IPC: FD = %d, condition = %s", source,
                      conditions_string)
-        
-        # Read a line from the file object
-        cmdline = sender.readline()
-        if not cmdline:             # Empty line means end of file
-            # close the IPC pipes
-            sender.close()
-            reply.close()
-            
-            # Stop calling this function
-            return False
-        
-        logger.debug(u"IPC command: %r", cmdline)
-        
-        # Parse and act on command
-        cmd, args = cmdline.rstrip(u"\r\n").split(None, 1)
-        
-        if cmd == u"NOTFOUND":
-            fpr, address = args.split(None, 1)
-            logger.warning(u"Client not found for fingerprint: %s, ad"
-                           u"dress: %s", fpr, address)
-            if self.use_dbus:
-                # Emit D-Bus signal
-                mandos_dbus_service.ClientNotFound(fpr, address)
-        elif cmd == u"DISABLED":
-            for client in self.clients:
-                if client.name == args:
-                    logger.warning(u"Client %s is disabled", args)
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.Rejected()
-                    break
-            else:
-                logger.error(u"Unknown client %s is disabled", args)
-        elif cmd == u"SENDING":
-            for client in self.clients:
-                if client.name == args:
-                    logger.info(u"Sending secret to %s", client.name)
-                    client.checked_ok()
-                    if self.use_dbus:
-                        # Emit D-Bus signal
-                        client.GotSecret()
-                    break
-            else:
-                logger.error(u"Sending secret to unknown client %s",
-                             args)
-        elif cmd == u"GETATTR":
-            attr_name, fpr = args.split(None, 1)
-            for client in self.clients:
-                if client.fingerprint == fpr:
-                    attr_value = getattr(client, attr_name, None)
-                    logger.debug("IPC reply: %r", attr_value)
-                    pickle.dump(attr_value, reply)
-                    break
-            else:
-                logger.error(u"Client %s on address %s requesting "
-                             u"attribute %s not found", fpr, address,
-                             attr_name)
-                pickle.dump(None, reply)
-        else:
-            logger.error(u"Unknown IPC command: %r", cmdline)
-        
-        # Keep calling this function
+
+        # error or the other end of multiprocessing.Pipe has closed
+        if condition & gobject.IO_HUP or condition & gobject.IO_ERR:
+            return False
+        
+        # Read a request from the child
+        request = parent_pipe.recv()
+        logger.debug(u"IPC request: %s", repr(request))
+        command = request[0]
+        
+        if command == 'init':
+            fpr = request[1]
+            address = request[2]
+            
+            for c in self.clients:
+                if c.fingerprint == fpr:
+                    client = c
+                    break
+            else:
+                logger.warning(u"Client not found for fingerprint: %s, ad"
+                               u"dress: %s", fpr, address)
+                if self.use_dbus:
+                    # Emit D-Bus signal
+                    mandos_dbus_service.ClientNotFound(fpr, address)
+                parent_pipe.send(False)
+                return False
+            
+            gobject.io_add_watch(parent_pipe.fileno(),
+                                 gobject.IO_IN | gobject.IO_HUP,
+                                 functools.partial(self.handle_ipc,
+                                                   parent_pipe = parent_pipe,
+                                                   client_object = client))
+            parent_pipe.send(True)
+            # remove the old hook in favor of the new above hook on same fileno
+            return False
+        if command == 'funcall':
+            funcname = request[1]
+            args = request[2]
+            kwargs = request[3]
+            
+            parent_pipe.send(('data', getattr(client_object, funcname)(*args, **kwargs)))
+
+        if command == 'getattr':
+            attrname = request[1]
+            if callable(client_object.__getattribute__(attrname)):
+                parent_pipe.send(('function',))
+            else:
+                parent_pipe.send(('data', client_object.__getattribute__(attrname)))
+
+        if command == 'setattr':
+            attrname = request[1]
+            value = request[2]
+            setattr(client_object, attrname, value)
+            
         return True
 
 
@@ -1553,6 +1690,8 @@
                         u"interval": u"5m",
                         u"checker": u"fping -q -- %%(host)s",
                         u"host": u"",
+                        u"approved_delay": u"5m",
+                        u"approved_duration": u"1s",
                         }
     client_config = configparser.SafeConfigParser(client_defaults)
     client_config.read(os.path.join(server_settings[u"configdir"],
@@ -1631,13 +1770,29 @@
     if server_settings["interface"]:
         service.interface = (if_nametoindex
                              (str(server_settings[u"interface"])))
+
+    global multiprocessing_manager
+    multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
         client_class = functools.partial(ClientDBus, bus = bus)
+    def client_config_items(config, section):
+        special_settings = {
+            "approved_by_default":
+                lambda: config.getboolean(section,
+                                          "approved_by_default"),
+            }
+        for name, value in config.items(section):
+            try:
+                yield (name, special_settings[name]())
+            except KeyError:
+                yield (name, value)
+    
     tcp_server.clients.update(set(
             client_class(name = section,
-                         config= dict(client_config.items(section)))
+                         config= dict(client_config_items(
+                        client_config, section)))
             for section in client_config.sections()))
     if not tcp_server.clients:
         logger.warning(u"No clients defined")

=== modified file 'mandos-monitor'
--- mandos-monitor	2009-12-25 23:13:47 +0000
+++ mandos-monitor	2010-09-02 18:53:38 +0000
@@ -23,11 +23,14 @@
 
 locale.setlocale(locale.LC_ALL, u'')
 
+import logging
+logging.getLogger('dbus.proxies').setLevel(logging.CRITICAL)
+
 # Some useful constants
 domain = 'se.bsnet.fukt'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
-version = "1.0.14"
+version = "1.0.15"
 
 # Always run in monochrome mode
 urwid.curses_display.curses.has_colors = lambda : False
@@ -50,12 +53,12 @@
                                      self.property_changed,
                                      client_interface,
                                      byte_arrays=True)
-
+        
         self.properties.update(
             self.proxy.GetAll(client_interface,
                               dbus_interface = dbus.PROPERTIES_IFACE))
         super(MandosClientPropertyCache, self).__init__(
-            proxy_object=proxy_object, *args, **kwargs)
+            *args, **kwargs)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -101,6 +104,10 @@
                                      self.got_secret,
                                      client_interface,
                                      byte_arrays=True)
+        self.proxy.connect_to_signal(u"NeedApproval",
+                                     self.need_approval,
+                                     client_interface,
+                                     byte_arrays=True)
         self.proxy.connect_to_signal(u"Rejected",
                                      self.rejected,
                                      client_interface,
@@ -108,9 +115,9 @@
     
     def checker_completed(self, exitstatus, condition, command):
         if exitstatus == 0:
-            self.logger(u'Checker for client %s (command "%s")'
-                        u' was successful'
-                        % (self.properties[u"name"], command))
+            #self.logger(u'Checker for client %s (command "%s")'
+            #            u' was successful'
+            #            % (self.properties[u"name"], command))
             return
         if os.WIFEXITED(condition):
             self.logger(u'Checker for client %s (command "%s")'
@@ -131,16 +138,25 @@
         self.logger(u'Checker for client %s completed mysteriously')
     
     def checker_started(self, command):
-        self.logger(u'Client %s started checker "%s"'
-                    % (self.properties[u"name"], unicode(command)))
+        #self.logger(u'Client %s started checker "%s"'
+        #            % (self.properties[u"name"], unicode(command)))
+        pass
     
     def got_secret(self):
         self.logger(u'Client %s received its secret'
                     % self.properties[u"name"])
     
-    def rejected(self):
-        self.logger(u'Client %s was rejected'
-                    % self.properties[u"name"])
+    def need_approval(self, timeout, default):
+        if not default:
+            message = u'Client %s needs approval within %s seconds'
+        else:
+            message = u'Client %s will get its secret in %s seconds'
+        self.logger(message
+                    % (self.properties[u"name"], timeout/1000))
+    
+    def rejected(self, reason):
+        self.logger(u'Client %s was rejected; reason: %s'
+                    % (self.properties[u"name"], reason))
     
     def selectable(self):
         """Make this a "selectable" widget.
@@ -225,6 +241,10 @@
 #             self.proxy.unpause()
 #         elif key == u"RET":
 #             self.open()
+        elif key == u"+":
+            self.proxy.Approve(True)
+        elif key == u"-":
+            self.proxy.Approve(False)
         else:
             return key
     

=== added file 'plugins.d/plymouth'
Binary files plugins.d/plymouth and plugins.d/plymouth differ
=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2010-03-27 18:39:02 +0000
+++ plugins.d/usplash.c	2010-09-01 13:04:23 +0000
@@ -42,8 +42,8 @@
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
@@ -607,6 +607,10 @@
     
     setsid();
     ret = chdir("/");
+    if(ret == -1){
+      perror("chdir");
+      _exit(EX_OSERR);
+    }
 /*     if(fork() != 0){ */
 /*       _exit(EXIT_SUCCESS); */
 /*     } */

